#ifndef EXPR_H
#define EXPR_H

#include <iostream>
#include <memory>
#include <string>

class Expr {
 public:
  virtual std::string toString() const { return "<expr>"; }

  friend std::ostream& operator<<(std::ostream& out, const Expr& expr) {
    return out << expr.toString();
  }

  virtual int eval() = 0;

  // This method allows implementing deep copy (no sharing)
  virtual std::unique_ptr<Expr> copy() const = 0;

  virtual ~Expr() {}
};

struct Number : public Expr {
  int value;

  Number(int value) : value(value) {}

  virtual int eval() override { return value; }

  std::string toString() const override { return std::to_string(value); }

  std::unique_ptr<Expr> copy() const override {
    return std::unique_ptr<Expr>(new Number(*this));
  }
};

struct BinaryExpr : public Expr {
 protected:
  std::unique_ptr<Expr> lhs;
  char op;
  std::unique_ptr<Expr> rhs;

 public:
  BinaryExpr(Expr& lhs, char op, Expr& rhs)
      : lhs(lhs.copy()), op(op), rhs(rhs.copy()) {}

  std::string toString() const override {
    return "(" + lhs->toString() + op + rhs->toString() + ")";
  }

  Expr& left() { return *lhs; }

  BinaryExpr(const BinaryExpr& that)
      : lhs(that.lhs->copy()), op(that.op), rhs(that.rhs->copy()) {}

  BinaryExpr& operator=(const BinaryExpr& that) = delete;
};

struct Add final : public BinaryExpr {
  Add(Expr& lhs, Expr& rhs) : BinaryExpr(lhs, '+', rhs) {}

  virtual int eval() override { return this->lhs->eval() + this->rhs->eval(); }

  std::unique_ptr<Expr> copy() const override {
    // return std::unique_ptr<Expr>(new Add(*lhs, *rhs));
    return std::make_unique<Add>(*lhs, *rhs);
  }
};

struct Mul final : public BinaryExpr {
  Mul(Expr& lhs, Expr& rhs) : BinaryExpr(lhs, '*', rhs) {}

  virtual int eval() override { return lhs->eval() * rhs->eval(); }

  std::unique_ptr<Expr> copy() const override {
    return std::unique_ptr<Expr>(new Mul(*lhs, *rhs));
  }
};

#endif  // EXPR_H
