#ifndef SMART_PTR_H
#define SMART_PTR_H

template<class T>
T&& Move(T& x) {
  return static_cast<T&&>(x);
}

// Note: Watch the CppCon video in the lecture notes
//
// For Monday:
// - vector moving example
// - lvalues vs rvalues
// - meaning of std::move
// - implemeting std::unique_ptr
// - advantages: no destructors, no overhead
// - the strategy below
//
// 1. use a local variable
// 2. use unique_ptr
// 3. use shared_ptr
//    - no cycles
//    - break cycles with weak_ptr or T*
// 4. roll your own smart pointer

// a pointer type that cannot be copied
template<class T>
class UniquePtr final {
  T* ptr;
public:
  // Takes ownership of the pointer ptr
  UniquePtr(T * ptr) : ptr(ptr) {}
  UniquePtr(UniquePtr const& other) = delete;

  UniquePtr(UniquePtr&& other)
    : ptr(other.ptr) {
    other.ptr = nullptr;
  }

  UniquePtr& operator=(UniquePtr const& other) = delete;

  UniquePtr& operator=(UniquePtr&& other) {
    if (this == &other) return *this;
    
    delete ptr;
    ptr = other.ptr;
    other.ptr = nullptr;

    return *this;
  }

  ~UniquePtr() { delete ptr; }

  // pointer operations, we delegate them to the pointer member

  T& operator *() const { return *ptr; }

  T& operator *() { return *ptr; }

  T* operator ->() const { return ptr; }

  T* operator ->() { return ptr; }
};

// For Wednesday: A simple shared pointer, no array support
template<class T>
class SharedPtr final {
private:
  // This is the actual data structure all shared pointers point to
  struct Cell {
    size_t count;
    T value;
  };
  
  Cell * cell;

  // Private constructor for creating a SharedPtr from a Cell
  SharedPtr(Cell * cell) : cell(cell) {}

public:
  // Default constructor
  SharedPtr() : cell(nullptr) {}
  
  // Copy constructor
  SharedPtr(SharedPtr& that) : cell(that.cell) {
    if (cell) {
      cell->count++;
    }
  }

  // Copy assignment operator
  SharedPtr& operator = (SharedPtr& that) {
    if (cell && --cell->count == 0) {
      delete cell;
    }
    cell = that.cell;
    if (cell) {
      cell->count++;
    }

    return *this;
  }

  SharedPtr& operator = (SharedPtr && that) {
    if (cell && --cell->count == 0) {
      delete cell;
    }
    cell = that.cell;
    that.cell = nullptr;
    
    return *this;
  }

  // Extract the value
  T& operator *() {
    return cell->value;
  }

  T* operator ->() {
    return &cell->value;
  }

  // Const versions of the above
  const T& operator *() const {
    return cell->value;
  }

  const T* operator ->() const {
    return &cell->value;
  }

  // Check if this pointer is null
  operator bool() const {
    return cell;
  }

  // Destructor
  ~SharedPtr() {
    if (cell && --cell->count == 0) {
      delete cell;
    }
  }

  // Static function to make a SharedPtr
  //
  // The "..." part allows us to pass values to the constructor of T.
  //
  // We will get to the "&&" part later, std::forward makes sure that we are
  // passing things by reference or value correctly.
  //
  // This is beyond CS 32
  template<class... Args>
  static SharedPtr Make(Args&&... args) {
    return SharedPtr(new Cell{
        .count = 1,
        .value = T(std::forward<Args>(args)...),
      });
  }
};

#endif // SMART_PTR_H
